/*

	mplayerplug-in

	This is a plugin for Netscape and Mozilla that uses mplayer to play supported videoes from websites
	It handles embedded and full page files equally.

	License: GPL

	Created 2003-01-09

	version 1.0	2003-12-02
	version 0.95	2003-10-01
	Version 0.91	2003-09-10
	Version 0.90	2003-09-07

	Authors:
	Kevin DeKorte	kdekorte@yahoo.com

 	Contributors
	Fabian Franz	gmplayer patches and support
	Todd Kirby	embed scale aspect option
	Ed Fisher	novop option
	Don Estberg	post processing options and QT Streams

	Code inspired by the webfreetv.com plugin http://www.webfreetv.com/linux/ by markus.rechberger@bluebull.com

 *
 */

#include "mplayerplug-in.h"

// This routine is only called when the plugin library is newer than the pluginrc file

char *NPP_GetMIMEDescription(void)
{

    char *ret;
    char MimeTypes[4000], line[1000], parse[1000], buffer[1000],
	codecname[32], customline[255];
    FILE *config, *customtypes;
    int i, use_custom_mime_types;


    enable_real = 0;
    enable_qt = 1;
    enable_wm = 1;
    enable_mpeg = 1;
    use_custom_mime_types = 0;

    DEBUG = 1;
    if (DEBUG)
	printf("Reading config file for codecs\n");
    // clear the buffer
    for (i = 0; i < 4000; i++) {
	MimeTypes[i] = '\0';
    }

    config = NULL;

    if (config == NULL) {
	snprintf(buffer, 1000, "%s", getenv("HOME"));
	strlcat(buffer, "/.mplayer/mplayerplug-in.conf", 1000);
	config = fopen(buffer, "r");
    }

    if (config == NULL) {
	snprintf(buffer, 1000, "%s", getenv("HOME"));
	strlcat(buffer, "/.mozilla/mplayerplug-in.conf", 1000);
	config = fopen(buffer, "r");
    }

    if (config == NULL) {
	config = fopen("/etc/mplayerplug-in.conf", "r");
    }

    if (config == NULL) {
	// no config file
    } else {
	while (fgets(buffer, 1000, config) != NULL) {

	    if (DEBUG)
		printf("Buffer: %s\n", buffer);

	    if (strncasecmp(buffer, "enable-qt", 9) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &enable_qt);
		if (DEBUG)
		    printf("qt:%i\n", enable_qt);
		continue;
	    }
	    // Windows Media
	    if (strncasecmp(buffer, "enable-wm", 9) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &enable_wm);
		if (DEBUG)
		    printf("wm:%i\n", enable_wm);
		continue;
	    }
	    // RealMedia
	    if (strncasecmp(buffer, "enable-real", 11) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &enable_real);
		if (DEBUG)
		    printf("real:%i\n", enable_real);
		continue;
	    }
	    // MPEG
	    if (strncasecmp(buffer, "enable-mpeg", 11) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &enable_mpeg);
		if (DEBUG)
		    printf("mpeg:%i\n", enable_mpeg);
		continue;
	    }

	    if (strncasecmp(buffer, "use-mimetypes", 13) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &use_custom_mime_types);
		if (DEBUG)
		    printf("custom mimetypes:%i\n", use_custom_mime_types);
		continue;
	    }


	}
	fclose(config);
    }

    if (use_custom_mime_types == 1) {
	customtypes = NULL;

	if (customtypes == NULL) {
	    snprintf(buffer, 1000, "%s", getenv("HOME"));
	    strlcat(buffer, "/.mplayer/mplayerplug-in.types", 1000);
	    customtypes = fopen(buffer, "r");
	}

	if (customtypes == NULL) {
	    snprintf(buffer, 1000, "%s", getenv("HOME"));
	    strlcat(buffer, "/.mozilla/mplayerplug-in.types", 1000);
	    config = fopen(buffer, "r");
	}

	if (customtypes == NULL) {
	    customtypes = fopen("/etc/mplayerplug-in.types", "r");
	}

	if (customtypes != NULL) {
	    while (fgets(customline, sizeof(customline), customtypes)) {
		strlcat(MimeTypes, customline, sizeof(MimeTypes));
	    }
	    fclose(customtypes);
	}


    } else {

	if (enable_qt == 1) {
	    strlcat(MimeTypes,
		    "video/quicktime:mov:Quicktime;"
		    "video/x-quicktime:mov:Quicktime;"
		    "video/quicktime:mp4:Quicktime;"
		    "video/quicktime:sdp:Quicktime - Session Description Protocol;"
		    "application/x-quicktimeplayer:mov:Quicktime;",
		    sizeof(MimeTypes));
	}

	if (enable_wm == 1) {
	    strlcat(MimeTypes,
		    "video/x-ms-asf-plugin:asf,asx:Windows Media;"
		    "video/x-msvideo:avi:AVI;"
		    "video/msvideo:avi:AVI;"
		    "application/x-mplayer2:wmv:WMV;"
		    "video/x-ms-wm:asf:MSNBCPlayer;"
		    "video/x-ms-asf:asf,asx:Windows Media;"
		    "video/x-ms-wmv:wmv:Windows Media;"
		    "video/x-ms-wvx:wvx:Windows Media;"
		    "audio/x-ms-wax:wax:Windows Media;"
		    "audio/x-ms-wma:wma:Windows Media;"
		    "application/x-drm-v2:asx:Windows Media;"
		    "audio/wav:wav:Microsoft wave file;"
		    "audio/x-wav:wav:Microsoft wave file;",
		    sizeof(MimeTypes));
	}


	if (enable_real == 1) {
	    strlcat(MimeTypes,
		    "audio/x-pn-realaudio:ram,rm:RealAudio;"
		    "audio/x-pn-realaudio-plugin:rpm:RealAudio;"
		    "audio/x-realaudio:ra:RealAudio;", sizeof(MimeTypes));
	}

	if (enable_mpeg == 1) {
	    strlcat(MimeTypes,
		    "video/mpeg:mpg,mpeg:MPEG;"
		    "audio/mpeg:mpg,mpeg:MPEG;"
		    "video/x-mpeg:mpg,mpeg:MPEG;"
		    "video/x-mpeg2:mpv2,mp2ve:MPEG2;"
		    "audio/mpeg:mpg,mpeg:MPEG;"
		    "audio/x-mpeg:mpg,mpeg:MPEG;"
		    "audio/mpeg2:mp2:MPEG audio;"
		    "audio/x-mpeg2:mp2:MPEG audio;"
		    "audio/mpeg3:mp3:MPEG audio;"
		    "audio/x-mpeg3:mp3:MPEG audio;"
		    "audio/mp3:mp3:MPEG audio;"
		    "video/mp4:mp4:MPEG 4 Video;", sizeof(MimeTypes));
	}
	// FLI
	strlcat(MimeTypes,
		"video/fli:fli,flc:FLI animation;"
		"video/x-fli:fli,flc:FLI animation;", sizeof(MimeTypes));

	// Vivo
	strlcat(MimeTypes, "video/vnd.vivo:viv,vivo:VivoActive;",
		sizeof(MimeTypes));

    }
    ret = strdup(MimeTypes);
    if (DEBUG)
	printf("%s\n", ret);
    DEBUG = 0;
    return ret;

}

/* This function returns the decription to Mozilla */

NPError NPP_GetValue(void *future, NPPVariable variable, void *value)
{
    NPError err = NPERR_NO_ERROR;

    switch (variable) {
    case NPPVpluginNameString:
	/* have to have these names in the name string to make Atom Films work */
	*((const char **) value) =
	    "MPlayer Plug-in supports QuickTime Plug-in 6.0 and Windows Media Player Plugin";
	break;
    case NPPVpluginDescriptionString:
	*((const char **) value) =
	    "<a href=\"http://mplayerplug-in.sourceforge.net/\">mplayerplug-in</a> "
	    PACKAGE_VERSION
	    "<br><br>Video Player Plug-in for QuickTime and Windows Media Player streams using <a href=\"http://mplayerhq.hu\">MPlayer</a> <br>";
	break;
    default:
	err = NPERR_GENERIC_ERROR;
    }
    return err;
}

NPError NPP_Initialize(void)
{
    return NPERR_NO_ERROR;
}


jref NPP_GetJavaClass()
{
    return NULL;
}

void NPP_Shutdown(void)
{
}


NPError NPP_New(NPMIMEType pluginType,
		NPP instance,
		uint16 mode,
		int16 argc, char *argn[], char *argv[],
		NPSavedData * saved)
{
    PluginInstance *This;
    int i;
    char parse[1000];
    char logfile[1000];
    char *cp;

    if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

    instance->pdata = NPN_MemAlloc(sizeof(PluginInstance));
    This = (PluginInstance *) instance->pdata;
    InitPrivateData(instance);

    DESTROYED = 0;

    This->mode = mode;
    LoadConfigFile(instance);

    if (mode == NP_EMBED) {

	if (DEBUG)
	    fprintf(This->logfp, "Embedded mode\n");
	for (i = 0; i < argc; i++) {

	    if (DEBUG) {
		fprintf(This->logfp, "Argument Name: %s\n", argn[i]);
		fprintf(This->logfp, "Argument Value: %s\n", argv[i]);
	    }

	    if (strncasecmp(argn[i], "src", 3) == 0) {
		This->url = strdup(argv[i]);
		This->state = STATE_HAVEURL;
	    }

	    if (strncasecmp(argn[i], "filename", 8) == 0) {
		This->fname = strdup(argv[i]);
		This->state = STATE_HAVEURL;
		NPN_GetURL(instance, This->fname, NULL);
	    }

	    if ((strncasecmp(argn[i], "href", 4) == 0)
		|| (strncasecmp(argn[i], "qtsrc", 5) == 0)) {
		This->href = strdup(argv[i]);
		This->state = STATE_HAVEURL;
	    }

	    if (strncasecmp(argn[i], "height", 6) == 0) {
		sscanf(argv[i], "%i", &This->embed_height);
	    }

	    if (strncasecmp(argn[i], "width", 5) == 0) {
		sscanf(argv[i], "%i", &This->embed_width);
	    }

	    /* handle 'scale' attribute used by QT Plugin */
	    if (strncasecmp(argn[i], "scale", 5) == 0) {
		if (strncasecmp(argv[i], "aspect", 6) == 0) {
		    This->maintain_aspect = 1;
		}
	    }

	    if (strncasecmp(argn[i], "loop", 4) == 0) {

		lowercase(argv[i]);
		if (DEBUG)
		    fprintf(This->logfp, "argv[i]=%s\n", argv[i]);

		if (strstr(argv[i], "true") || strstr(argv[i], "yes")) {
		    This->loop = 1;
		} else {
		    This->loop = 0;
		}
		if (DEBUG)
		    fprintf(This->logfp, "loop=%i\n", This->loop);
	    }

	    if (This->nQtNext < 256
		&& (strncasecmp(argn[i], "qtnext", 6) == 0)
		&& argv[i][0] == '<') {
		snprintf(parse, 1000, "%s", strtok(&argv[i][1], ">"));
		if ((cp = strchr(parse, ' '))
		    && strlen(parse) == cp - parse + 1)
		    *cp = (char) NULL;
		This->qtNext[This->nQtNext++] = strdup(parse);
		snprintf(parse, 1000, "%s", strtok(NULL, "<"));
		if (strcmp(parse, "T")) {
		    if (DEBUG)
			fprintf(This->logfp,
				"qtNext%i expected \"T\" found \"%s\"\n",
				This->nQtNext, parse);
		    This->nQtNext--;
		} else {
		    snprintf(parse, 1000, "%s", strtok(NULL, ">\n"));
		    if (strcmp(parse, "myself")) {
			if (DEBUG)
			    fprintf(This->logfp,
				    "qtNext%i expected \"myself\" found \"%s\"\n",
				    This->nQtNext, parse);
			This->nQtNext--;
		    } else if (DEBUG)
			fprintf(This->logfp, "qtNext%i=%s\n",
				This->nQtNext,
				This->qtNext[This->nQtNext - 1]);
		}
	    }



	}
    } else {
	if (DEBUG)
	    fprintf(This->logfp, "New, full mode %i\n", This->mode);
    }


    if (This != NULL)
	return NPERR_NO_ERROR;
    else
	return NPERR_OUT_OF_MEMORY_ERROR;
}


NPError NPP_Destroy(NPP instance, NPSavedData ** save)
{

    char cmd[1024];
    PluginInstance *This;
    int ret;

    if (DEBUG)
	printf("destroy %p\n", instance);

    if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

    This = (PluginInstance *) instance->pdata;
    pthread_mutex_lock(&(This->playlist_mutex));
    DESTROYED = 1;
    if (This != NULL) {
	if (DEBUG)
	    fprintf(This->logfp, "\nIn NPP_Destroy\n");

	if (DEBUG)
	    ret = fclose(This->logfp);
    }
    /* cleanup streams and files */
    pthread_mutex_unlock(&(This->playlist_mutex));
    pthread_attr_destroy(&(This->thread_attr));
    pthread_mutex_destroy(&(This->playlist_mutex));
    pthread_cond_destroy(&(This->playlist_complete_cond));

    if (This != NULL) {
	if (DEBUG)
	    printf("Freeing memory\n");

	if (This->td != NULL) {
	    if (DEBUG)
		printf("This->td = %p\n", This->td);
	    NPN_MemFree(This->td);
	    This->td = NULL;
	    if (DEBUG)
		printf("This->td freed\n");
	}
	NPN_MemFree(instance->pdata);
	This = NULL;
	instance->pdata = NULL;
	if (DEBUG)
	    printf("This freed\n");
    }

    return NPERR_NO_ERROR;

}



NPError NPP_SetWindow(NPP instance, NPWindow * window)
{
    PluginInstance *This;
    Widget netscape_widget, screen;
    NPSetWindowCallbackStruct *ws;
    int ret, i;
    char *buffer, *buffer2;
    Node *n;

    if (instance == NULL) {
	if (DEBUG)
	    printf("SetWindow Called with invalid instance\n");
	return NPERR_INVALID_INSTANCE_ERROR;
    }

    This = (PluginInstance *) instance->pdata;

    if ((window == NULL) || (window->window == NULL)) {
	return NPERR_NO_ERROR;
    }


    if (DEBUG) {
	fprintf(This->logfp, "Set Window \nNPP %p\n", instance);
	fprintf(This->logfp, "This %p\n", This);
	fprintf(This->logfp, "This->state %i\n", This->state);
	printf("SW %p\n", instance);
    }


    if (This->window != (Window) window->window) {
	if (DEBUG)
	    fprintf(This->logfp,
		    "Hey we got a new window! old: 0x%x    new 0x%x\n",
		    This->window, window->window);
    }

    if (This->state < STATE_WINDOWSET) {
	if (DEBUG) {
	    fprintf(This->logfp, "Size: %d %d %x\n", window->x, window->y,
		    window->window);
	    fprintf(This->logfp, "Size: %dx%d \n", window->width,
		    window->height, window->window);
	}

	ws = (NPSetWindowCallbackStruct *) window->ws_info;

	netscape_widget =
	    XtWindowToWidget((Display *) ws->display,
			     (Window) window->window);

	XtAddCallback(netscape_widget, XtNdestroyCallback, DestroyCB,
		      (XtPointer) This);
	XtAddEventHandler(netscape_widget, ExposureMask, FALSE,
			  (XtEventHandler) RedrawCB, This);

	This->display = ws->display;
	This->window = (Window) window->window;
	This->netscape_widget = netscape_widget;
	This->width = window->width;
	This->height = window->height;
	This->state = STATE_WINDOWSET;


    } else {
	if (This->setwindow == 0) {
	    if (DEBUG)
		fprintf(This->logfp, "back in SetWindow\n");
	    if (DEBUG)
		fprintf(This->logfp, "New Size: %ix%i\n", window->width,
			window->height);
	    if (This->mode == NP_EMBED) {
		This->embed_width = window->width;
		This->embed_height = window->height;
	    }
	    This->setwindow++;

	    if (This->state != STATE_GETTING_PLAYLIST) {
		This->state = STATE_GETTING_PLAYLIST;
		if (This->href != NULL) {
		    if (isMms(This->href)) {
			snprintf(This->td->list->url, 1024, "%s",
				 This->href);
			This->state = STATE_STARTED_PLAYER;
			This->td->list->mmsstream = 1;
			Redraw(This->netscape_widget, instance, NULL);
		    }
		} else {
		    if (This->url != NULL) {
			if (isMms(This->url)) {
			    snprintf(This->td->list->url, 1024, "%s",
				     This->url);
			    This->state = STATE_STARTED_PLAYER;
			    This->td->list->mmsstream = 1;
			    Redraw(This->netscape_widget, instance, NULL);
			}
		    }
		}
	    }
	}
    }

    if (DEBUG) {
	fprintf(This->logfp, "Exiting SetWindow\n");
	fflush(This->logfp);
    }

    return NPERR_NO_ERROR;
}

void
NPP_URLNotify(NPP instance, const char *url, NPReason reason,
	      void *notifyData)
{
    PluginInstance *This;
    int playable;
    Node *n;

    if (instance == NULL)
	return;

    This = (PluginInstance *) instance->pdata;

    if (DEBUG)
	printf("NPN_URLNotify called\n");


    if (reason == NPRES_DONE) {
	playable = 0;
	pthread_mutex_lock(&(This->playlist_mutex));

	n = This->td->list;
	while (n != NULL) {
	    if (n->play == 1)
		playable = 1;
	    n = n->next;
	}

	if (playable) {
	    if (This->threadsignaled == 0) {
		pthread_cond_signal(&(This->playlist_complete_cond));
		This->threadsignaled = 1;
	    }
	}

	pthread_mutex_unlock(&(This->playlist_mutex));
    }
}

NPError
NPP_NewStream(NPP instance,
	      NPMIMEType type,
	      NPStream * stream, NPBool seekable, uint16 * stype)
{
    NPByteRange range;
    PluginInstance *This;
    Node *n;

    if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

    This = (PluginInstance *) instance->pdata;

    if (DEBUG) {
	fprintf(This->logfp, "NewStream Requested\n");
	printf("New Stream requested state: %i\n", This->state);
	printf("URL: %s\n", stream->url);
    }

    if (This->mode == NP_FULL) {
	This->url = strdup(stream->url);
    }

    if (This->state <= STATE_GETTING_PLAYLIST) {
	This->state = STATE_GETTING_PLAYLIST;
	Redraw(This->netscape_widget, instance, NULL);
    }
    if (DEBUG)
	fprintf(This->logfp, "NP_NORMAL\n");
    *stype = NP_NORMAL;

    return NPERR_NO_ERROR;

}


/* PLUGIN DEVELOPERS:
 *	These next 2 functions are directly relevant in a plug-in which
 *	handles the data in a streaming manner. If you want zero bytes
 *	because no buffer space is YET available, return 0. As long as
 *	the stream has not been written to the plugin, Navigator will
 *	continue trying to send bytes.  If the plugin doesn't want them,
 *	just return some large number from NPP_WriteReady(), and
 *	ignore them in NPP_Write().  For a NP_ASFILE stream, they are
 *	still called but can safely be ignored using this strategy.
 */

int32 STREAMBUFSIZE = 0X0FFFFFFF;	/* If we are reading from a file in NPAsFile
					 * mode so we can take any size stream in our
					 * write call (since we ignore it) */
int32 NPP_WriteReady(NPP instance, NPStream * stream)
{
    PluginInstance *This;
    Node *n;

    if (instance == NULL)
	return -1;

    if (DESTROYED)
	return -1;
    This = (PluginInstance *) instance->pdata;
    if (This->state == STATE_PLAY_CANCELLED)
	return -1;

    if (0) {
	fprintf(This->logfp, "in WriteReady. stream->url: %s\n",
		stream->url);
	fprintf(This->logfp, "in WriteReady, state = %i\n", This->state);
	fflush(This->logfp);

	fprintf(This->logfp, "Setting of This->forced = %i\n",
		This->forced);
    }

    if (This == NULL)
	return -1;
    if (This->td == NULL)
	return -1;

    pthread_mutex_lock(&(This->playlist_mutex));
    if (strlen(stream->url) >= 1023)
	return -1;
    n = This->td->list;
    while (n != NULL) {
	if (strlen(n->url) != 0) {
	    if (URLcmp(n->url, stream->url) == 0) {
		break;
	    } else {
		if (strstr(stream->url, n->url) != NULL) {
		    break;
		}
	    }
	} else {
	    snprintf(n->url, 1024, "%s", stream->url);
	    break;
	}
	n = n->next;
    }

    if (n != NULL) {
	if (strlen(n->fname) == 0) {
	    snprintf(n->fname, 1024, "%s",
		     tempnam("/tmp", "mplayerplug-inXXXXXX"));
	    if (DEBUG)
		printf("WR tempname: %s\n", n->fname);

	}
	pthread_mutex_unlock(&(This->playlist_mutex));
	return STREAMBUFSIZE;
    } else {
	if (DEBUG)
	    printf("didn't find the node in the playlist\n %s\n",
		   stream->url);
	n = newNode();
	snprintf(n->url, 1024, "%s", stream->url);
	snprintf(n->fname, 1024, "%s",
		 tempnam("/tmp", "mplayerplug-inXXXXXX"));
//      if ((This->href != NULL) && (This->url != NULL)){
//          if (URLcmp(This->url, stream->url) == 0) {
//              n->play = 0;
//          }
//      }
	addToEnd(This->td->list, n);
	pthread_mutex_unlock(&(This->playlist_mutex));
	return STREAMBUFSIZE;
    }
//    }
}


int32
NPP_Write(NPP instance, NPStream * stream, int32 offset, int32 len,
	  void *buffer)
{
    PluginInstance *This;
    int ret;
    FILE *fp;
    Node *n, *hrefnode;
    char message[1024];

    if (instance == NULL)
	return -1;

    if (DESTROYED)
	return -1;
    This = (PluginInstance *) instance->pdata;
    if (This->state == STATE_PLAY_CANCELLED)
	return -1;

    if (0)
	fprintf(This->logfp, "in Write. stream->url: %s\n", stream->url);
    if (0)
	fprintf(This->logfp, "in Write, state = %i\n", This->state);
    if (0)
	fflush(This->logfp);

    if (This == NULL)
	return -1;
    if (This->td == NULL)
	return -1;

    pthread_mutex_lock(&(This->playlist_mutex));
    // find the node that matches the URL and open the file name in it.
    if (strlen(stream->url) >= 1023)
	return -1;
    n = This->td->list;
    while (n != NULL) {
	if (URLcmp(n->url, stream->url) == 0) {
	    break;
	} else {
	    if (strstr(stream->url, n->url) != NULL) {
		break;
	    }
	}
	n = n->next;
    }

    if (n == NULL) {
	pthread_mutex_unlock(&(This->playlist_mutex));
	return -1;
    } else {
    	// if we have determined that we are not going to play the file
	// stop downloading it, and clean it up during cleanup
        if ((n->play == 0) && (n->playlist == 0) && (n->bytes > 4096)) {
		n->delete = 1;
		return -1;
	}

	if (n->status != STATE_CANCELLED) {
	    fp = fopen(n->fname, "a+");
	    fflush(stdout);
	    fseek(fp, offset, SEEK_SET);
	    ret = fwrite(buffer, 1, len, fp);
	    n->bytes = n->bytes + ret;
	    snprintf(message, 1024, "Downloaded %li bytes", n->bytes);
	    if (This->state < STATE_PLAYING)
		DrawUI(This->netscape_widget, This->td->instance, message);
	    n->retrieved = 1;
	    fclose(fp);

	    if (This->href != NULL) {
		if (This->hrefrequested == 0) {
		    if (DEBUG)
			printf("requesting href\n");
		    This->hrefrequested = 1;
		    n = This->td->list;
		    while (n != NULL) {
			if (URLcmp(n->url, This->href) == 0) {
			    break;
			} else {
			    if (strstr(stream->url, This->href) != NULL) {
				break;
			    }
			}
			n = n->next;
		    }
		    if (n == NULL) {
			if (DEBUG)
			    printf("href not in list\n");
			hrefnode = newNode();
			strlcpy(hrefnode->url, This->href,
				sizeof(hrefnode->url));
			addToEnd(This->td->list, hrefnode);
		    } else {
			if (DEBUG)
			    printf("href in list\n");
		    }
		    NPN_GetURL(instance, This->href, NULL);
		    pthread_mutex_unlock(&(This->playlist_mutex));
		    return ret;
		}
	    }


	    if (n->status != STATE_DOWNLOADED_ENOUGH) {
		if (This->baseurl == NULL)
		    This->baseurl = getURLBase(n->url);

		if (DEBUG)
		    printf("calling buildPlaylist with filename %s\n",
			   n->fname);
		if (!isMms(n->url)) {
		    buildPlaylist(instance, n->fname, n);
		} else {
		    n->mmsstream = 1;
		}
	    }
	    if (n->play == 1) {
		if (n->mmsstream == 1) {
		    if (This->threadsignaled == 0) {
			if (DEBUG)
			    printf("signalling player from write\n");
			pthread_cond_signal(&
					    (This->
					     playlist_complete_cond));
			This->threadsignaled = 1;
		    }
		} else {
		    //if (This->mode != NP_FULL) {
		    if (n->bytes > (This->cachesize * 1024)) {
			if (This->threadsignaled == 0) {
			    if (DEBUG)
				printf("signalling player from write\n");
			    pthread_cond_signal(&
						(This->
						 playlist_complete_cond));
			    This->threadsignaled = 1;
			}
		    }
		    //}
		}
	    }

	    n->status = STATE_DOWNLOADED_ENOUGH;
	    pthread_mutex_unlock(&(This->playlist_mutex));
	    return ret;
	}
    }
}


NPError NPP_DestroyStream(NPP instance, NPStream * stream, NPError reason)
{
    PluginInstance *This;
    int playable, all_retrieved, all_above_cache;
    Node *n;
    Node *closest;
    int ret;

    if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;
    This = (PluginInstance *) instance->pdata;

    if (DEBUG)
	fprintf(This->logfp, "in destroy stream\n");
    if (DEBUG)
	fflush(This->logfp);

    if (DEBUG)
	printf("NPP_DestroyStream called %i\n URL: %s\n", reason,
	       stream->url);


    if (reason == NPRES_DONE) {
	playable = 0;
	pthread_mutex_lock(&(This->playlist_mutex));
	if (strlen(stream->url) >= 1023)
	    return;
	n = This->td->list;
	while (n != NULL) {
	    if (URLcmp(n->url, stream->url) == 0) {
		if (DEBUG)
		    printf("Destroy stream found a URL match\n");
		break;
	    } else {
		if (strstr(stream->url, n->url) != NULL) {
		    break;
		}
	    }
	    n = n->next;
	}

	if (n != NULL) {
	    n->retrieved = 1;

	    // adjust playlist if we have speed set on the nodes
	    closest = NULL;
	    n = This->td->list;
	    while (n != NULL) {
		if ((n->speed > 0) && (n->play == 1)) {
		    if (closest == NULL) {
			closest = n;
		    } else {
			if (abs(This->qt_speed - closest->speed) >
			    abs(This->qt_speed - n->speed)) {
			    // this node has a closer speed to preferred
			    closest->play = 0;
			    closest = n;
			} else {
			    // the previous node had a closer speed
			    n->play = 0;
			}
		    }
		    break;
		}
		n = n->next;
	    }

	    // test for all retrieved
	    if (This->threadsignaled == 0) {
		all_retrieved = 1;
		n = This->td->list;
		while (n != NULL) {
		    if (n->retrieved == 0) {
			all_retrieved = 0;
			break;
		    }
		    n = n->next;
		}
		// if all retrieved signal player
		if (all_retrieved) {
		    if (This->threadsignaled == 0) {
			if (DEBUG)
			    printf("signalling player\n");
			pthread_cond_signal(&
					    (This->
					     playlist_complete_cond));
			This->threadsignaled = 1;
		    }
		}
	    }
	    // if not then see if all not retrieved are above cache,
	    if (This->threadsignaled == 0) {
		all_above_cache = 1;
		n = This->td->list;
		while (n != NULL) {
		    if (n->bytes < (This->cachesize * 1024)) {
			all_above_cache = 0;
			break;
		    }
		    n = n->next;
		}
		// if all above cache signal player
		if (all_above_cache) {
		    if (This->threadsignaled == 0) {
			if (DEBUG)
			    printf("signalling player\n");
			pthread_cond_signal(&
					    (This->
					     playlist_complete_cond));
			This->threadsignaled = 1;
		    }
		}
	    }

	    if (This->threadsignaled == 0) {
		// look for mmsstream
		n = This->td->list;
		while (n != NULL) {
		    if (n->mmsstream) {
			if (This->threadsignaled == 0) {
			    if (DEBUG)
				printf("signalling player\n");
			    pthread_cond_signal(&
						(This->
						 playlist_complete_cond));
			    This->threadsignaled = 1;
			}
			break;
		    }
		    n = n->next;
		}
	    }
	}
	pthread_mutex_unlock(&(This->playlist_mutex));
    }

    return NPERR_NO_ERROR;
}


void NPP_StreamAsFile(NPP instance, NPStream * stream, const char *fname)
{
    PluginInstance *This;
    FILE *infile;
    char c;
    char s[MAX_BUF_LEN];
    int i = 0;

    if (instance == NULL)
	return;

    This = (PluginInstance *) instance->pdata;

    if (DEBUG)
	printf("SaF %p\n", instance);


    if (DEBUG) {
	fprintf(This->logfp, "in stream as file. stream->url: %s\n",
		stream->url);
	fprintf(This->logfp, "in stream as file. fname: %s\n", fname);
	printf("in stream as file. stream->url: %s\n", stream->url);
	printf("in stream as file. fname: %s\n", fname);
    }
    This->td = (ThreadData *) NPN_MemAlloc(sizeof(ThreadData));
    This->td->list = newNode();
    This->td->instance = instance;
    This->url = strdup(stream->url);
    This->fname = strdup(fname);
    snprintf(This->td->list->url, 1024, "%s", This->url);
    snprintf(This->td->list->fname, 1024, "%s", This->fname);
    if (!isMms(This->url)) {
	buildPlaylist(instance, This->fname, This->td->list);
    } else {
	This->td->list->mmsstream = 1;
    }

    Redraw(This->netscape_widget, instance, NULL);
    return;
}


void NPP_Print(NPP instance, NPPrint * printInfo)
{
    return;
}

static void DestroyCB(Widget widget, XtPointer client_data,
		      XtPointer call_data)
{
    PluginInstance *This = (PluginInstance *) client_data;
    char cmd[1000], buffer[1024];
    int count, tries;
    FILE *ps;

    if (DEBUG)
	printf("window destroy callback\n");
    if (widget == This->netscape_widget) {
	if (This == NULL)
	    return;
	DESTROYED = 1;
	sendCommand(This->fifoname, "quit");
	if (DEBUG)
	    printf("sent quit\n");

	pthread_mutex_lock(&(This->playlist_mutex));
	XtRemoveCallback(This->netscape_widget, XtNdestroyCallback,
			 DestroyCB, (XtPointer) This);
	XtRemoveEventHandler(This->netscape_widget, ExposureMask, FALSE,
			     (XtEventHandler) RedrawCB, This);
	This->netscape_widget = NULL;
	This->state = STATE_PLAY_CANCELLED;

	if (DEBUG)
	    printf("starting wait loop\n");
	tries = 0;
	do {
	    snprintf(cmd, 1000,
		     "ps -eaf | grep mplayer | grep 0x%x | wc -l",
		     This->window);
	    ps = popen(cmd, "r");
	    if (ps != NULL) {
		while (fgets(buffer, 1024, ps) != NULL) {
		    if (DEBUG)
			printf("buffer %s\n", buffer);
		    sscanf(buffer, "%i", &count);
		    if (DEBUG)
			printf("waiting for mplayer to close count = %i\n",
			       count);
		}
		pclose(ps);
	    }
	    // if mplayer has not shutdown after 10 tries... kill it
	    tries++;
	    if (tries > 10) {
		snprintf(cmd, 1000,
			 "ps -eaf | grep mplayer | grep 0x%x | awk '{print $2}' | xargs kill -9",
			 This->window);
		ps = popen(cmd, "r");
		if (ps != NULL) {
		    pclose(ps);
		}
	    }
	} while (count > 1);

	remove(This->fifoname);
	deleteList(This->td->list);
	This->td->list == NULL;

	if (This->td != NULL) {
	    if (DEBUG)
		printf("This->td = %p\n", This->td);
	    NPN_MemFree(This->td);
	    This->td = NULL;
	    if (DEBUG)
		printf("This->td freed\n");
	}
	if (This->baseurl != NULL) {
	    NPN_MemFree(This->baseurl);
	    This->baseurl = NULL;
	    if (DEBUG)
		printf("This->baseurl freed\n");
	}
	if (This->hostname != NULL) {
	    NPN_MemFree(This->hostname);
	    This->hostname = NULL;
	    if (DEBUG)
		printf("This->hostname freed\n");
	}


	This->td = (ThreadData *) NPN_MemAlloc(sizeof(ThreadData));
	This->td->list = newNode();
	This->td->instance = NULL;
	pthread_mutex_unlock(&(This->playlist_mutex));

    } else {
	if (DEBUG)
	    printf("Widgets don't match %i != %i\n", widget,
		   This->netscape_widget);
    }

}

static void RedrawCB(Widget widget, XtPointer client_data,
		     XtPointer call_data)
{
    char message[1024];

    PluginInstance *This = (PluginInstance *) client_data;
    if (DEBUG)
	printf("redrawing window\n");

    if (DESTROYED || This->noredraw)
	return;

    if (This != NULL) {
	if (This->lastmessage != NULL) {
	    if (strlen(This->lastmessage) != 0) {
		if (This->td->instance != NULL) {
		    if (This->state >= STATE_GETTING_PLAYLIST) {
			strcpy(message, This->lastmessage);
			DrawUI(This->netscape_widget, This->td->instance,
			       message);
		    }
		}
	    }
	}
    }
}

void Redraw(Widget w, NPP instance, XEvent * event)
{

    PluginInstance *This;
    Node *node;
    int i = 0;
    char xval[32], yval[32], yval2[32];
    char *filename;

    if (DEBUG)
	printf("Redraw\n");
    if (instance == NULL) {
	if (DEBUG)
	    printf("Our instance is null!\n");
	return;
    }
    // structures

    This = (PluginInstance *) instance->pdata;
    if (This->redrawcalled == 1)
	return;

    This->redrawcalled = 1;
    This->td->instance = instance;
    This->td->w = w;

    if (This->download == 1) {
	DrawUI(w, instance, "Downloading Media...");
    } else {
	DrawUI(w, instance, "Loading Media...");
    }


    if (This->href) {
	snprintf(This->td->list->url, 1024, "%s", This->href);
    } else if (This->fname) {
	snprintf(This->td->list->url, 1024, "%s", This->fname);
    } else {
	snprintf(This->td->list->url, 1024, "%s", This->url);
    }

    if (This->mode == NP_FULL) {
	snprintf(xval, 32, "%i", This->width);
	snprintf(yval, 32, "%i", This->height);
	snprintf(yval2, 32, "%i", This->height - 16);
	if (This->fileonly_embed == 0)
	    This->noembed = 1;
    } else {
	snprintf(xval, 32, "%i", This->embed_width);
	snprintf(yval, 32, "%i", This->embed_height);
	snprintf(yval2, 32, "%i", This->embed_height - 16);
    }

    This->baseurl = getURLBase(This->td->list->url);
    This->hostname = getURLHostname(This->td->list->url);
    if (This->keep_download == 1) {
	This->td->list->delete = 0;
	filename = getURLFilename(This->td->list->url);
	snprintf(This->td->list->fname, 1024, "%s/%s", This->download_dir,
		 filename);
	if (filename)
	    NPN_MemFree(filename);
    } else {
	snprintf(This->td->list->fname, 0, "");
    }

    snprintf(This->td->argv[i++], 1024, "%s", This->player);

    if (This->novop == 1) {
	snprintf(This->td->argv[i++], 1024, "-vop");
	snprintf(This->td->argv[i++], 1024, "null");
    } else {
	if (This->vopopt != NULL) {
	    snprintf(This->td->argv[i++], 1024, "-vop");
	    snprintf(This->td->argv[i++], 1024, "%s", This->vopopt);
	    if (strncmp(This->vopopt, "scale=", 6) == 0) {
		snprintf(This->td->argv[i++], 1024, "-fs");
	    }
	}
    }

    if (This->noembed == 0) {
	if (This->window != 0) {
	    snprintf(This->td->argv[i++], 1024, "-wid");
	    snprintf(This->td->argv[i++], 1024, "0x%x", This->window);
	}
    }

    if (DEBUG)
	printf("window defined %s \n", This->td->argv[i - 1]);

    if ((This->embed_width == 0)
	&& (This->mode == NP_EMBED)) {

    } else {
	if (This->scaling == 1) {
	    if (This->use_gmplayer > 0) {
		snprintf(This->td->argv[i++], 1024, "-screenw");
		snprintf(This->td->argv[i++], 1024, "%s", xval);
		snprintf(This->td->argv[i++], 1024, "-screenh");
		snprintf(This->td->argv[i++], 1024, "%s", yval);
		if (This->maintain_aspect == 1) {
		    snprintf(This->td->argv[i++], 1024, "-xy");
		    snprintf(This->td->argv[i++], 1024, "%s", xval);
		} else {
		    snprintf(This->td->argv[i++], 1024, "-x");
		    snprintf(This->td->argv[i++], 1024, "%s", xval);
		}
		if (This->use_gmplayer == 2) {
		    if (This->maintain_aspect == 0) {
			snprintf(This->td->argv[i++], 1024, "-y");
			snprintf(This->td->argv[i++], 1024, "%s", yval2);
		    }
		    snprintf(This->td->argv[i++], 1024, "-skin");
		    snprintf(This->td->argv[i++], 1024, "mini");
		    if (This->noembed == 0) {
			snprintf(This->td->argv[i++], 1024, "-guiwid");
			snprintf(This->td->argv[i++], 1024, "0x%x",
				 This->window);
		    }
		} else {
		    if (This->maintain_aspect == 0) {
			snprintf(This->td->argv[i++], 1024, "-y");
			snprintf(This->td->argv[i++], 1024, "%s", yval);
		    }
		}
	    } else {
		if (This->maintain_aspect == 1) {
		    snprintf(This->td->argv[i++], 1024, "-xy");
		    snprintf(This->td->argv[i++], 1024, "%s", xval);
		} else {
		    snprintf(This->td->argv[i++], 1024, "-x");
		    snprintf(This->td->argv[i++], 1024, "%s", xval);
		    snprintf(This->td->argv[i++], 1024, "-y");
		    snprintf(This->td->argv[i++], 1024, "%s", yval);
		}
	    }
	}
    }

    if (This->rtsp_use_tcp == 1) {
	snprintf(This->td->argv[i++], 1024, "-rtsp-stream-over-tcp");
    }

    if (This->cachesize > 0) {
	snprintf(This->td->argv[i++], 1024, "-cache");
	snprintf(This->td->argv[i++], 1024, "%i", This->cachesize);
    } else {
	snprintf(This->td->argv[i++], 1024, "-nocache");
    }

    if (This->loop == 1) {
	snprintf(This->td->argv[i++], 1024, "-loop");
	snprintf(This->td->argv[i++], 1024, "0");
    }

    if (This->vo) {
	snprintf(This->td->argv[i++], 1024, "-vo");
	snprintf(This->td->argv[i], 1024, "%s", This->vo);
	if ((strcmp(This->td->argv[i], "x11") == 0)
	    || (strstr(This->td->argv[i], ",x11") != NULL)) {
	    i++;
	    snprintf(This->td->argv[i++], 1024, "-zoom");	/* -vo x11 needs -zoom for rescaling */
	} else {
	    i++;
	}
    }

    if (This->ao) {
	snprintf(This->td->argv[i++], 1024, "-ao");
	snprintf(This->td->argv[i++], 1024, "%s", This->ao);
    }

    if (This->output_display) {
	snprintf(This->td->argv[i++], 1024, "-display");
	snprintf(This->td->argv[i++], 1024, "%s", This->output_display);
    }

    snprintf(This->td->argv[i++], 1024, "-osdlevel");
    snprintf(This->td->argv[i++], 1024, "%i", This->osdlevel);

    /* some extra flags for mplayer */
    snprintf(This->td->argv[i++], 1024, "-nojoystick");	/* annoying... */
    snprintf(This->td->argv[i++], 1024, "-nofs");	/* no full screen */
    snprintf(This->td->argv[i++], 1024, "-slave");	/* no full screen */

    // if we don't do this next part, using use-gui=[yes|mini] has annoying popups
    if (strcasecmp(This->player, "gmplayer") == 0) {
	snprintf(This->td->argv[i++], 1024, "-really-quiet");
    } else {
	snprintf(This->td->argv[i++], 1024, "-quiet");
    }
    snprintf(This->td->argv[i++], 1024, "");

    // ok we have the command line setup
    if (DEBUG)
	printf("ready to setup threads\n");

    pthread_attr_init(&(This->thread_attr));
    pthread_attr_setdetachstate(&(This->thread_attr),
				PTHREAD_CREATE_DETACHED);
    pthread_cond_init(&(This->playlist_complete_cond), NULL);
    This->current_node = NULL;


    if (DEBUG)
	printf("creating thread - NP_EMBED\n");
    DrawUI(w, instance, "Getting playlist...");
    if (This->state < STATE_GETTING_PLAYLIST)
	This->state = STATE_GETTING_PLAYLIST;
    if (DEBUG)
	printf("creating player thread\n");
    pthread_create(&(This->player_thread), &(This->thread_attr),
		   playPlaylist, (void *) This->td);
    // recommended slight pause
    usleep(1);

    if (DEBUG)
	printf("MAIN THREAD DONE\n");
}
