/*

	mplayerplug-in

	License: GPL

	This file is the header for the mplayerplug-in.c

*/

#define _XOPEN_SOURCE 500
#define _GNU_SOURCE

#include <pthread.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <sys/stat.h>

#include "../config.h"
#include "npapi.h"
#include "missing.h"

extern int errno;
int DEBUG;
int enable_real;
int enable_qt;
int enable_wm;
int enable_mpeg;
int DESTROYED;

#define MAX_BUF_LEN 255
#define STATE_RESET 0
#define STATE_NEW 1
#define STATE_HAVEURL 3
#define STATE_WINDOWSET 4
#define STATE_READY 5
#define STATE_QUEUED 6
#define STATE_DOWNLOADING 7
#define STATE_DOWNLOADED_ENOUGH 8		

//#define STATE_PLAYING 10
#define STATE_CANCELLED 11

#define STATE_NEWINSTANCE 100
#define STATE_GETTING_PLAYLIST 110
#define STATE_STARTED_PLAYER 115
#define STATE_PLAYLIST_COMPLETE 120
#define STATE_PLAYLIST_NEXT 125
#define STATE_PLAYING 130
#define STATE_PLAY_COMPLETE 140
#define STATE_PLAY_CANCELLED 150

typedef struct lnode {
    char url[1024];
    char fname[1024];
    int status;
    int retrieved;
    int play;
    int speed;
    int playlist;
    int mmsstream;
    int delete;
    long int bytes;
    struct lnode *next;
} Node;

typedef struct _ThreadData {
    Widget w;
    NPP instance;
    char argv[50][1024];
    Node *list;
} ThreadData;

typedef struct _PluginInstance {
    int state;
    char *url;
    char *fname;
    char *href;
    uint16 mode;
    Widget netscape_widget;
    Display *display;
    Window window;
    uint32 width;
    uint32 height;
    uint32 embed_width;
    uint32 embed_height;
    FILE *logfp;
    char *logfile;
    char *output_display;
    int cachesize;
    int use_gmplayer;
    int qt_speed;
    char *player;
    int setwindow;
    int loop;
    int maintain_aspect;
    int scaling;
    int fileonly_embed;
    int rtsp_use_tcp;
    char *vo;
    char *ao;
    char *vopopt;
    int noembed;
    int novop;
    int osdlevel;
    int shmkey;
    uint32 nQtNext;
    char *qtNext[256];
    int download;
    char *download_dir;
    int keep_download;
    char *local_fname;
    int forced;
    Node *current_node;
    ThreadData *td;
    char *baseurl;
    char *hostname;
    int redrawcalled;
    int threadsignaled;
    int hrefrequested;
    char *lastmessage;
    pthread_t player_thread;
    pthread_attr_t thread_attr;
    pthread_cond_t playlist_complete_cond;
    pthread_mutex_t playlist_mutex;
    char *fifoname;
    int noredraw;
} PluginInstance;


void Redraw(Widget w, NPP instance, XEvent * event);
void CloseEvents(Widget w, XtPointer instance, XAnyEvent * event,
		 Boolean * b);
static void DestroyCB(Widget widget, XtPointer client_data,
		      XtPointer call_data);
static void RedrawCB(Widget widget, XtPointer client_data,
		     XtPointer call_data);
extern void lowercase(char string[]);
extern void LoadConfigFile(NPP instance);
extern void InitPrivateData(NPP instance);
extern void DrawUI(Widget w, NPP instance, char *message);
extern void DrawProgress(Widget w, NPP instance, int percent);
extern void SendKey(PluginInstance * This, Window window, char key);
extern void SendCmd(int fp, char *cmd);
extern void buildPlaylist(NPP instance, char *url, Node * parent);
extern Node *newNode();
extern void *playPlaylist(void *td);
extern void buildPlaylist(NPP instance, char *url, Node * parent);
extern void deleteList(Node * l);
extern void addToEnd(Node * l, Node * newnode);
extern char *getURLBase(char *url);
extern char *getURLHostname(char *url);
extern char *getURLFilename(char *url);
extern int isMms(char *url);
extern void sendCommand(char *fname, char *command);
extern int URLcmp(const char *url1, const char *url2);
