/*

	mplayerplug-in

	License: GPL

	This file contains support routines for the mplayerplug-in

*/


#include "mplayerplug-in.h"

void lowercase(char string[])
{
    int i = 0;

    while (string[i]) {
	string[i] = tolower(string[i]);
	i++;
    }

    return;
}


void InitPrivateData(NPP instance)
{

    PluginInstance *This;
    int i;

    This = (PluginInstance *) instance->pdata;

    This->state = STATE_NEW;
    This->url = NULL;
    This->fname = NULL;
    This->href = NULL;
    This->netscape_widget = NULL;
    This->display = NULL;
    This->window = (Window) NULL;
    This->width = 0;
    This->height = 0;
    This->embed_height = 0;
    This->embed_width = 0;
    This->setwindow = 0;
    This->forced = 0;
    This->nQtNext = 0;
    for (i = 0; i < 256; i++)
	This->qtNext[i] = NULL;

    This->cachesize = 512;
    This->logfile = NULL;
    This->logfp = NULL;
    This->player = strdup("mplayer");
    This->use_gmplayer = 0;
    This->qt_speed = 6;
    This->vo = NULL;
    This->ao = NULL;
    This->noembed = 0;
    This->novop = 0;
    This->osdlevel = 0;
    This->vopopt = NULL;
    This->maintain_aspect = 1;
    This->scaling = 1;
    This->loop = 0;
    This->fileonly_embed = 0;
    This->rtsp_use_tcp = 0;
    This->output_display = NULL;
    This->download = 0;
    This->download_dir = strdup(getenv("HOME"));
    This->local_fname = NULL;
    This->keep_download = 0;
    This->td = NULL;
    This->baseurl = NULL;
    This->hostname = NULL;
    This->td = (ThreadData *) NPN_MemAlloc(sizeof(ThreadData));
    This->td->list = newNode();
    This->td->instance = NULL;
    This->redrawcalled = 0;
    This->threadsignaled = 0;
    This->hrefrequested = 0;
    This->lastmessage = (char *) NPN_MemAlloc(sizeof(char) * 1024);
    pthread_mutex_init(&(This->playlist_mutex), NULL);
    This->fifoname = tempnam("/tmp", "fifoXXXXXX");
    This->noredraw = 0;
}

void LoadConfigFile(NPP instance)
{

    PluginInstance *This;
    int i, j;
    FILE *config;
    char buffer[1000];
    char parse[1000];
    char logfile[1000];
    char *cp;

    This = (PluginInstance *) instance->pdata;

    // load config file

    config = NULL;

    if (config == NULL) {
	snprintf(buffer, 1000, "%s", getenv("HOME"));
	strlcat(buffer, "/.mplayer/mplayerplug-in.conf", 1000);
	config = fopen(buffer, "r");
    }

    if (config == NULL) {
	snprintf(buffer, 1000, "%s", getenv("HOME"));
	strlcat(buffer, "/.mozilla/mplayerplug-in.conf", 1000);
	config = fopen(buffer, "r");
    }

    if (config == NULL) {
	config = fopen("/etc/mplayerplug-in.conf", "r");
    }

    if (config == NULL) {
	// no config file
    } else {
	while (fgets(buffer, sizeof(buffer), config) != NULL) {
	    if (strncasecmp(buffer, "cachesize", 9) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &This->cachesize);
		if (This->cachesize < 0)
		    This->cachesize = 0;
		if (This->cachesize > 65535)
		    This->cachesize = 65535;
		continue;
	    }

	    if (strncasecmp(buffer, "debug", 5) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &DEBUG);
		if (DEBUG != 0)
		    DEBUG = 1;
		continue;
	    }

	    if (strncasecmp(buffer, "noembed", 7) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &This->noembed);
		if (This->noembed != 0)
		    This->noembed = 1;
		continue;
	    }

	    if (strncasecmp(buffer, "novop", 5) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &This->novop);
		if (This->novop != 0)
		    This->novop = 1;
		continue;
	    }

	    if (strncasecmp(buffer, "vopopt", 6) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "\n"));
		This->novop = 0;
		This->vopopt = strdup(parse);
		continue;
	    }

	    if (strncasecmp(buffer, "prefer-aspect", 13) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &This->maintain_aspect);
		if (This->maintain_aspect != 0)
		    This->maintain_aspect = 1;
		continue;
	    }

	    if (strncasecmp(buffer, "scaling", 7) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &This->scaling);
		if (This->scaling != 0)
		    This->scaling = 1;
		continue;
	    }

	    if (strncasecmp(buffer, "fileonly-embed", 14) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &This->fileonly_embed);
		if (This->fileonly_embed != 0)
		    This->fileonly_embed = 1;
		continue;
	    }

	    if (strncasecmp(buffer, "rtsp-use-tcp", 12) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &This->rtsp_use_tcp);
		if (This->rtsp_use_tcp != 0)
		    This->rtsp_use_tcp = 1;
		continue;
	    }

	    if (strncasecmp(buffer, "logfile", 7) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(logfile, 1000, "%s", strtok(NULL, "=\n"));
		if (strstr(logfile, "$HOME") != NULL) {
		    snprintf(buffer, sizeof(buffer),
			     "%s%s", getenv("HOME"), logfile + 5);
		    strlcpy(logfile, buffer, sizeof(logfile));
		}
		This->logfile = strdup(logfile);
		This->logfp = fopen(This->logfile, "a");
		continue;
	    }

	    if (strncasecmp(buffer, "qt-speed", 8) == 0) {
		sprintf(parse, "%s", strtok(buffer, "="));
		sprintf(parse, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &This->qt_speed);
		printf("QT Speed: %i\n", This->qt_speed);
		continue;
	    }

	    if (strncasecmp(buffer, "use-gui", 7) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "=\n"));
		if (strncasecmp(parse, "yes", 3) == 0) {
		    This->use_gmplayer = 1;
		} else {
		    if (strncasecmp(parse, "mini", 4) == 0) {
			This->use_gmplayer = 2;
		    }
		}

		if (This->use_gmplayer > 0) {
		    if (This->player)
			free(This->player);
		    This->player = strdup("gmplayer");
		}
		continue;
	    }

	    if (strncasecmp(buffer, "vo", 2) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "=\n"));
		This->vo = strdup(parse);
		continue;
	    }

	    if (strncasecmp(buffer, "ao", 2) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "=\n"));
		This->ao = strdup(parse);
		continue;
	    }

	    if (strncasecmp(buffer, "display", 7) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "=\n"));
		This->output_display = strdup(parse);
		continue;
	    }

	    if (strncasecmp(buffer, "download", 8) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &This->download);
		if (This->download != 0)
		    This->download = 1;
		continue;
	    }

	    if (strncasecmp(buffer, "dload-dir", 9) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "=\n"));
		if (strstr(parse, "$HOME") != NULL) {
		    snprintf(buffer, sizeof(buffer), "%s%s",
			     getenv("HOME"), parse + 5);
		    strlcpy(parse, buffer, sizeof(parse));
		}
		if (This->download_dir != NULL)
		    free(This->download_dir);
		This->download_dir = strdup(parse);
		continue;
	    }

	    if (strncasecmp(buffer, "keep-download", 13) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &This->keep_download);
		if (This->keep_download != 0)
		    This->keep_download = 1;
		continue;
	    }


	    if (strncasecmp(buffer, "osdlevel", 8) == 0) {
		snprintf(parse, 1000, "%s", strtok(buffer, "="));
		snprintf(parse, 1000, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &This->osdlevel);
		if (This->osdlevel < 0)
		    This->osdlevel = 0;
		if (This->osdlevel > 3)
		    This->osdlevel = 3;
	    }

	}
    }

    if (This->logfp == NULL && DEBUG == 1)
	DEBUG = 0;
    if (This->download_dir == NULL && This->download == 1)
	This->download_dir = strdup(getenv("HOME"));

}

char *getURLHostname(char *url)
{

    char *hostname;
    char *loc;
    int i;

    if (DEBUG)
	printf("entering getURLHostname\n");
    if (url == NULL)
	return NULL;
    if (strlen(url) == 0)
	return NULL;
    hostname = NPN_MemAlloc(sizeof(char) * (strlen(url) + 1));
    strcpy(hostname, url);

    loc = strstr(url, "://");
    if (loc == NULL) {
	NPN_MemFree(hostname);
	return NULL;
    }
    loc = loc + 3;

    i = 0;

    while (loc[0] != '/') {
	hostname[i] = loc[0];
	loc++;
	i++;
	if (i > strlen(url)) {
	    i = 0;
	    break;
	}
    }
    hostname[i] = '\0';
    if (DEBUG)
	printf("exiting getURLHostname with %s\n", hostname);

    return hostname;

}

char *getURLFilename(char *url)
{

    char *filename;
    char *tmp;

    if (url == NULL)
	return NULL;
    if (strlen(url) == 0)
	return NULL;
    filename = NPN_MemAlloc(sizeof(char) * (strlen(url) + 1));
    tmp = rindex(url, '/');
    if (tmp == NULL) {
	strcpy(filename, url);
	return filename;
    }
    strcpy(filename, tmp);
    return filename;
}

char *getURLBase(char *url)
{

    char *base;
    int i;

    if (DEBUG)
	printf("in getURLBase\n");
    if (url == NULL)
	return NULL;
    if (strlen(url) == 0)
	return NULL;
    base = NPN_MemAlloc(sizeof(char) * (strlen(url) + 1));
    strcpy(base, url);

    for (i = strlen(url) - 1; i > 0; i--) {
	if (base[i] != '/') {
	    base[i] = '\0';
	} else {
	    break;
	}
    }
    if (strlen(base) == 0) {
	NPN_MemFree(base);
	base = NULL;
    }
    if (DEBUG)
	printf("exiting URL base with %s\n", base);

    return base;
}

int isMms(char *url)
{
    if ((strncmp(url, "mms://", 6) == 0)
	|| (strncmp(url, "mmst://", 7) == 0)
	|| (strncmp(url, "rtsp://", 7) == 0)) {
	return 1;
    } else {
	return 0;
    }
}

void mmsToHttp(char *dest, char *src)
{
    char *temp;

    temp = (char *) NPN_MemAlloc(sizeof(char) * (strlen(src) + 1 +
						 sizeof("http") -
						 sizeof("mms")));
    strcpy(temp, "http");
    strcat(temp, src + 3);
    strcpy(dest, temp);
    NPN_MemFree(temp);
    return;
}

void sendCommand(char *fname, char *command)
{
    FILE *fifo;

    fifo = fopen(fname, "w");
    if (fifo != NULL) {
	fprintf(fifo, "%s\n", command);
	fclose(fifo);
    }
}

int URLcmp(const char *url1, const char *url2)
{
    char *buffer1, *buffer2;
    char *tmp;

    if (strcmp(url1, url2) == 0)
	return 0;

    // replace %20 with spaces in both strings
    buffer1 = strdup(url1);
    buffer2 = strdup(url2);

    while (tmp = strstr(buffer1, "%20")) {
	if (tmp != NULL) {
	    *tmp = ' ';
	    tmp++;
	    *tmp = '\0';
	    strcat(buffer1, tmp + 2);
	}
    }

    while (tmp = strstr(buffer2, "%20")) {
	if (tmp != NULL) {
	    *tmp = ' ';
	    tmp++;
	    *tmp = '\0';
	    strcat(buffer2, tmp + 2);
	}
    }

    if (strcmp(buffer1, buffer2) == 0) {
	//if (DEBUG) printf("URL's match\n buffer1 %s len:%i\n buffer2 %s len %i\n",buffer1,strlen(buffer1),buffer2,strlen(buffer2));
	free(buffer1);
	free(buffer2);
	return 0;
    } else {
	//if (DEBUG) printf("URL's don't match\n buffer1 %s len:%i\n buffer2 %s len%i\n",buffer1,strlen(buffer1),buffer2,strlen(buffer2));
	free(buffer1);
	free(buffer2);
	return -1;
    }
}

// Linked List methods
// This is how the playlist is going to be managed

void initialize(Node * l)
{
    l = NULL;
}

Node *newNode()
{
    Node *n = NPN_MemAlloc(sizeof(Node));
    n->url[0] = '\0';
    n->fname[0] = '\0';
    n->next = NULL;
    n->bytes = 0;
    n->play = 1;
    n->playlist = 0;
    n->speed = -1;
    n->mmsstream = 0;
    n->delete = 1;
    n->status = STATE_NEWINSTANCE;
    n->retrieved = 0;
    n->next = NULL;
    return n;
}

void deleteNode(Node * n)
{
    if (n != NULL) {
	n->url[0] = '\0';
	n->fname[0] = '\0';
	n->bytes = 0;
	n->play = 1;
	n->mmsstream = 0;
	n->delete = 1;
	n->status = STATE_NEWINSTANCE;
	n->retrieved = 0;
	NPN_MemFree(n);
	n = NULL;
    }
}

void copyNode(Node * dest, Node * src)
{
    if (src != NULL && dest != NULL) {
	strlcpy(dest->url, src->url, sizeof(dest->url));
	strlcpy(dest->fname, src->fname, sizeof(src->fname));
	dest->bytes = src->bytes;
	dest->play = src->play;
	dest->mmsstream = src->mmsstream;
	dest->playlist = src->playlist;
	dest->speed = dest->speed;
	dest->delete = src->delete;
    }
}

void deleteList(Node * l)
{
    Node *oldList;

    if (DEBUG)
	printf("Entering deleteList\n");
    while (l != NULL) {
	oldList = l;
	if (DEBUG) {
	    printf("DELETE %s\n", l->url);
	    if (l->delete)
		printf("File delete %s\n", l->fname);
	    printf("size: %i\n", l->bytes);
	    printf("play: %i\n", l->play);
	    printf("playlist: %i\n", l->playlist);
	    printf("speed: %i\n", l->speed);
	}
	if ((l->delete == 1) || (l->delete == 0 && l->play == 0)) {
	    if (strlen(l->fname) != 0)
		remove(l->fname);
	}
	l = l->next;
	deleteNode(oldList);
    }

}

/* Inserts node newnode after node item */
void insertafter(Node * item, Node * newnode)
{
    if (item == NULL || newnode == NULL) {
	if (DEBUG)
	    printf("Cannot insert\n");
    } else {
	newnode->next = item->next;
	item->next = newnode;
    }
}

void addToEnd(Node * l, Node * newnode)
{

    if (l == NULL) {
	l = newnode;
    } else {
	while (l->next != NULL)
	    l = l->next;
	l->next = newnode;
    }

}

int fexists(char *file)
{

    FILE *fp;

    fp = fopen(file, "r");
    if (fp != NULL) {
	fclose(fp);
	return 1;
    } else {
	return 0;
    }

}


void addToList(NPP instance, char *item, Node * parent, int speed)
{
    Node *node = NULL;
    PluginInstance *This;
    Node *n;
    char tmpdir[1024];
    char *tmp;
    char *filename;
    char localitem[1024];

    This = (PluginInstance *) instance->pdata;

    if (!isMms(item)) {
	if (strncasecmp(item, "http", 4) != 0) {

	    if (item[0] != '/') {
		strlcpy(tmpdir, item, sizeof(tmpdir));	// reuse the buffer
		if (This->baseurl != NULL)
		    strlcpy(localitem, This->baseurl, 1024);
		strlcat(localitem, tmpdir, 1024);
	    } else {
		if (This->hostname != NULL) {
		    snprintf(tmpdir, sizeof(tmpdir), "http://%s/%s",
			     This->hostname, item);
		    strlcpy(localitem, tmpdir, 1024);
		}
	    }

	} else {
	    // if :80 is in the URL, cut it out
	    strlcpy(localitem, item, sizeof(localitem));
	    tmp = strstr(localitem, ":80");
	    if (tmp != NULL) {
		*tmp = '\0';
		strlcat(localitem, tmp + 3, 1024);
	    }
	}
    } else {
	strlcpy(localitem, item, 1024);
    }


    n = This->td->list;
    if (DEBUG)
	printf("BUILD - traversing playlist\n");
    while (n != NULL) {
	if (URLcmp(n->url, localitem) == 0) {
	    if (DEBUG)
		printf("found a match\n");
	    break;
	}
	n = n->next;
    }

    if (n == NULL) {
	// add it in
	if (DEBUG)
	    printf("adding url %s\n", localitem);
	node = newNode();
	strlcpy(node->url, localitem, sizeof(node->url));
	node->speed = speed;
	addToEnd(parent, node);
	parent->play = 0;
	if (isMms(item)) {
	    node->mmsstream = 1;
	} else {
	    filename = getURLFilename(item);
	    if (This->keep_download == 1 && filename != NULL) {
		snprintf(node->fname, sizeof(node->fname), "%s/%s",
			 This->download_dir, filename);
		if (DEBUG)
		    printf("Assigned name: %s\n", node->fname);
		node->delete = 0;
	    }
	    if (filename != NULL)
		NPN_MemFree(filename);
	    NPN_GetURL(instance, node->url, NULL);
	}
    }
}

void buildPlaylist(NPP instance, char *file, Node * parent)
{
    FILE *fp;
    char buffer[4096];
    char buffer_lower[4096];
    char url[1024];
    PluginInstance *This;
    int size, found;
    char *p, *sub, *refurl;
    int i;
    int code;
    char *c;
    int speed;
    char *c1, *c2;
    char smallbuff[10];

    This = (PluginInstance *) instance->pdata;

    fp = fopen(file, "r");

    found = 0;
    if (fp != NULL) {
	while (!feof(fp)) {
	    memset(buffer, 0, sizeof(buffer));
	    memset(buffer_lower, 0, sizeof(buffer));
	    memset(url, 0, sizeof(url));

	    size = fread(buffer, 1, sizeof(buffer) - 1, fp);
	    buffer[sizeof(buffer) - 1] = '\0';

	    if (DEBUG)
		printf("READ: %i\n%s\n", size, buffer);

	    // quicktime files have blocks of 4 characters..
	    // rdrf,XXXX,url ,XXXX,url ending with 00

	    if (memmem(buffer, size, "rdrf", 4)) {
		if (DEBUG)
		    printf("found QT\n");
		p = buffer;
		while (p != NULL) {
		    p = memmem(p, size - ((long) buffer - (long) p),
			       "rdrf", 4);
		    if (p == NULL)
			break;
		    code = (int) (p[15]);
		    p = p + 16;
		    refurl = p;
		    if (DEBUG) {
			printf("URL: %s\n", (refurl));
			printf("CODE: %d\n", code);
		    }
		    if (code == -93 || code == -91 || code == -89) {
			if (DEBUG)
			    printf("Skipped URL: %s\n", p);
		    } else {
			p = memmem(p, size - ((long) buffer - (long) p),
				   "rmdr", 4);
			p = p + 8;
			// have to AND with 0x7F to remove the sign bit, otherwise numbers are HUGE
			speed =
			    ((p[0] & 127) << 24) + ((p[1] & 127) << 16) +
			    ((p[2] & 127) << 8) + (p[3] & 127);
			speed = speed / 1024;	// should get it close to 0 - 12
			if (parent->speed > speed)
			    speed = parent->speed;
			parent->playlist = 1;
			addToList(instance, refurl, parent, speed);
			printf("ADDED URL: %s\n", (refurl));
			found = 1;
		    }
		}
		break;
	    }
	    // smil format files
	    if (strncasecmp(buffer, "<smil", 5) == 0) {
		if (DEBUG)
		    printf("found smil file\n");
		snprintf(buffer_lower, 4096, "%s", buffer);
		lowercase(buffer_lower);
		p = buffer_lower;
		// Handle SEQ tag (bitrate specification)
		while (p != NULL) {
		    speed = -1;
		    i = size - ((long) buffer_lower - (long) p);
		    if (i < 1)
			break;
		    c = memmem(p, i, "<seq", 4);
		    if (c != NULL) {

			p = memmem(c, i, "system-bitrate=", 15);
			if (p != NULL) {
			    p += 15;
			    c1 = strchr(p, '\"') + 1;
			    if (c1 == NULL)
				c1 = p;
			    c2 = strchr(c1, '\"');
			    if (c2 == NULL) {
				c2 = strchr(c2, ' ');
				if (c2 == NULL)
				    c2 = strchr(c1, '>');
			    }
			    memcpy((char *) &smallbuff[0], c1,
				   (int) (c2 - c1));
			    smallbuff[(int) (c2 - c1)] = 0;	// Null-terminate
			    sscanf(smallbuff, "%d", &speed);
			    speed = speed / 20000;	// Adjust speed to rough 0-12 range
			} else {
			    if (DEBUG)
				printf("Unhandled SEQ tag!\n");
			}
		    } else {
			if (DEBUG)
			    printf("No SEQ tag!\n");
		    }
		    p = memmem(p, i, "video src=", 10);
		    if (p == NULL)
			break;
		    p = p + 10;
		    p = strchr(p, '\"') + 1;
		    if (p != NULL) {
			i = (long) p - (long) buffer_lower;
			snprintf(url, 1024, "%s", buffer + i);
		    }
		    sub = strchr(url, '\"');
		    if (sub != NULL) {
			*sub = '\0';
			if (DEBUG)
			    printf("URL: %s\n", url);
			addToList(instance, url, parent, speed);
			p = p + strlen(url) + 1;
			found = 1;
		    }
		}
		break;
	    }
	    // some mp3 files start with ID3, not a reference file
	    // not all MP3 files have the ID tag
	    if (strncasecmp(buffer, "ID3", 3) == 0) {
		if (DEBUG)
		    printf("found MP3\n");
		break;
	    }
	    // asx files are XMLish
	    // <Ref href = "url" />
	    if (strncasecmp(buffer, "<asx", 4) == 0) {
		if (DEBUG)
		    printf("found ASX file\n");
		snprintf(buffer_lower, 4096, "%s", buffer);
		lowercase(buffer_lower);
		p = buffer_lower;
		while (p != NULL) {
		    i = size - ((long) buffer_lower - (long) p);
		    if (i < 1)
			break;
		    p = memmem(p, i, "href", 4);
		    if (p == NULL)
			break;
		    p = p + 4;
		    p = strchr(p, '\"') + 1;
		    if (p != NULL) {
			i = (long) p - (long) buffer_lower;
			snprintf(url, 1024, "%s", buffer + i);
		    }
		    sub = strchr(url, '\"');
		    if (sub != NULL) {
			*sub = '\0';
			if (DEBUG)
			    printf("URL: %s\n", url);
			addToList(instance, url, parent, -1);
			p = p + strlen(url) + 1;
			found = 1;
		    }
		}
		break;
	    }
	    // asf files
	    if (strncasecmp(buffer, "[reference]", 11) == 0) {
		if (DEBUG)
		    printf("found ASF file\n");
		snprintf(buffer_lower, 4096, "%s", buffer);
		lowercase(buffer_lower);
		p = buffer_lower;
		while (p != NULL) {
		    i = size - ((long) buffer_lower - (long) p);
		    if (i < 1)
			break;
		    p = memmem(p, i, "ref", 3);
		    if (p == NULL)
			break;
		    p = p + 3;
		    p = strchr(p, '=') + 1;
		    if (p != NULL) {
			i = (long) p - (long) buffer_lower;
			snprintf(url, 1024, "%s", buffer + i);
		    }
		    sub = strchr(url, '\r');
		    if (sub == NULL)
			sub = strchr(url, '\n');
		    if (sub == NULL)
			sub = strchr(url, EOF);
		    if (sub != NULL) {
			*sub = '\0';
			if (DEBUG)
			    printf("URL: %s\n", url);
			parent->playlist = 1;
			addToList(instance, url, parent, -1);
			p = p + strlen(url) + 1;
			found = 1;
		    }
		}
		break;
	    }

	    if (!found)
		break;
	}
	fclose(fp);
    }
}

// Remove all single quotes from url to prevent remote execution
static void remove_quotes(char *url)
{
    char *p;

    // don't scan the string if it not in there, this should be a bit faster than the while loop

    p = strchr(url, '`');

    if (p == NULL) {
	return;
    } else {
	if (DEBUG) {
	    printf("single quotes in url (%s), truncating\n"
		   "WARNING: probably won't work! FIXME!\n", url);
	}
	*p = '\0';
    }
}

void *playPlaylist(void *td)
{

    FILE *player;
    char logfile[1024];
    char buffer[1024];
    char cmd[2048];
    char command[2048];
    char message[1024];
    char *eos;
    ThreadData *local_td;
    Node *local_list, *copy;
    int i;
    PluginInstance *This;
    Node *node;
    struct timeval now;
    struct timespec timeout;
    struct stat dload_stat;
    long int filesize;
    int ready, notfound;

    local_td = (ThreadData *) td;
    local_list = local_td->list;
    This = (PluginInstance *) local_td->instance->pdata;

    if (DEBUG)
	printf("playPlayList - deciding to wait\n");

    pthread_mutex_lock(&(This->playlist_mutex));
    if (This->state < STATE_STARTED_PLAYER) {
	pthread_cond_wait(&(This->playlist_complete_cond),
			  &(This->playlist_mutex));
	This->state = STATE_STARTED_PLAYER;
    }
    pthread_mutex_unlock(&(This->playlist_mutex));

    if (DEBUG)
	printf("playPlayList - waiting complete\n");


    DrawUI(local_td->w, local_td->instance, "Loading Media...");

    snprintf(cmd, 2048, "\0");
    i = 0;
    if (DEBUG)
	printf("building command string\n");
    while (strcmp(local_td->argv[i], "") != 0) {
	if (DEBUG)
	    printf("PLAY %i:%s\n", i, local_td->argv[i]);
	strlcat(cmd, local_td->argv[i], sizeof(cmd));
	strlcat(cmd, " ", sizeof(cmd));
	i++;
    }

    i = 0;
    while (This->qtNext[i] != NULL) {
	node = newNode();
	if (DEBUG)
	    printf("adding %s\n", This->qtNext[i]);
	strlcpy(node->url, This->qtNext[i], sizeof(node->url));
	i++;
	addToEnd(local_list, node);
    }

    // preview playlist
    if (DEBUG) {
	while (local_list != NULL) {
	    if (DEBUG)
		printf("PL URL: %s\n", local_list->url);
	    local_list = local_list->next;
	}
	local_list = local_td->list;
    }

    copy = newNode();
    pthread_mutex_lock(&(This->playlist_mutex));
    copyNode(copy, local_list);
    pthread_mutex_unlock(&(This->playlist_mutex));
    while (copy != NULL) {
	if (DEBUG)
	    printf("url: %s \n play %i\n", copy->url, copy->play);
	strlcpy(message, "Playing ", sizeof(message));
	strlcat(message, copy->url, sizeof(message));
	DrawUI(This->netscape_widget, local_td->instance, message);
	if (copy->play) {
	    strlcpy(command, cmd, sizeof(command));
	    // for rtsp streams we need to specify FPS
	    if (strncmp(copy->url, "rtsp", 4) == 0) {
		strlcat(command, " -fps ", sizeof(command));
		strlcat(command, " 30 ", sizeof(command));
	    }
	    if (copy->playlist == 1)
		strlcat(command, " -playlist ", sizeof(command));
	    strlcat(command, "\'", sizeof(command));

	    if (copy->mmsstream) {
		remove_quotes(copy->url);
		strlcat(command, copy->url, sizeof(command));
	    } else {
		if (strlen(copy->fname) == 0) {
		    pthread_mutex_lock(&(This->playlist_mutex));
		    local_list = local_list->next;
		    if (local_list == NULL) {
			deleteNode(copy);
			copy = NULL;
		    } else {
			copyNode(copy, local_list);
		    }
		    pthread_mutex_unlock(&(This->playlist_mutex));
		    continue;
		}
		strcat(command, copy->fname);
	    }
	    strlcat(command, "\' <> ", sizeof(command));
	    mkfifo(This->fifoname, S_IWUSR | S_IRUSR);
	    strlcat(command, This->fifoname, sizeof(command));

	    if (DEBUG) {
		printf("URL: %s\n", copy->url);
		printf("CMD: %s\n", command);
	    }

	    if (!copy->mmsstream) {
		stat(copy->fname, &dload_stat);
		do {
		    filesize = dload_stat.st_size;
		    sleep(1);
		    stat(copy->fname, &dload_stat);
		    if (DEBUG)
			printf("sleeping size:%i\n", dload_stat.st_size);
		    // if nothing has been downloaded then maybe the file is done and we should play
		    if (filesize == dload_stat.st_size)
			break;
		}
		while (dload_stat.st_size < (This->cachesize * 1024));
	    }

	    player = popen(command, "r");
	    This->state = STATE_PLAYING;
	    if (player != NULL) {
		while (fgets(buffer, sizeof(buffer), player) != NULL) {
		    if (DEBUG) {
			printf("READ: %s \n", buffer);
		    }
		    if (strstr(buffer, "Cache fill:") != NULL) {
			strlcpy(message, strstr(buffer, "Cache fill"),
				sizeof(message));
			notfound = 1;
			while (notfound) {
			    eos = strrchr(message, '\r');
			    if (eos == NULL)
				eos = message;
			    if (strstr(eos, "Cache fill:") == NULL) {
			    	// if we don't find "Cache fill:" then replace the \r with \0, making message shorter
				eos[0] = '\0';
				// break out of the loop if too short
				if (strlen(message) < 5)
				    break;
			    } else {
			    	// if we found the last one in the message string, move from \r to C in the string
			        eos++;
				notfound = 0;
			    }
			}
			// Only update the display if there is something worth displaying
			if (strstr(eos,"Cache fill:") != NULL)
			    DrawUI(This->netscape_widget,
				   local_td->instance, eos);
		    }
		    if (strstr(buffer, "Starting") != NULL) {
			strlcpy(message, "Playing ", sizeof(message));
			strlcat(message, copy->url, sizeof(message));
			DrawUI(This->netscape_widget, local_td->instance,
			       message);
		    }
		    if (strstr(buffer, "VO:") != NULL) {
			This->noredraw = 1;
		    }
		    if (strstr(buffer, "Quit") != NULL) {
			if (DEBUG)
			    printf("breaking read loop\n");
			break;
		    }
		}
		pclose(player);
		if (DEBUG)
		    printf("pclose done\n");
		// close the fifo
		remove(This->fifoname);
		if (DEBUG)
		    printf("fifo removed\n");
		This->state = STATE_PLAYLIST_NEXT;
	    }
	}
	if (DESTROYED)
	    return;
	This = (PluginInstance *) local_td->instance->pdata;
	if (This != NULL) {
	    if (DESTROYED)
		return;
	    if (This->state != STATE_PLAY_CANCELLED) {
		if (DESTROYED)
		    return;
		if (local_list != NULL) {
		    if (DESTROYED)
			return;
		    local_list = local_list->next;
		    if (local_list == NULL) {
			deleteNode(copy);
			copy = NULL;
		    } else {
			copyNode(copy, local_list);
		    }
		}
	    } else {
		break;
		copy = NULL;
	    }
	}
    }
}
