/*

	mplayerplug-in

	License: GPL

	This file contains User Interface routines for the mplayerplug-in

*/

#include "mplayerplug-in.h"
#include <X11/xpm.h>
#include "logo.xpm"

#define BORDER               10
#define LOGO_WIDTH          348
#define LOGO_HEIGHT          48

#ifndef HAVE_X11_XPM
    #error libXpm has not been found. Compilation cannot continue
#endif
static XFontStruct *font = NULL;
static Pixmap logo, logomask;

void Initialize(Display *dpy, Drawable d, GC gc)
{
    XpmCreatePixmapFromData(dpy, d, logo_xpm, &logo, &logomask, NULL);
    font = XQueryFont(dpy, XGContextFromGC(gc));
}

void DrawLogo(Display *dpy, Drawable d, GC gc) {
    XGCValues values;

    values.clip_mask = logomask;
    values.clip_x_origin = 0;
    values.clip_y_origin = 0;
    XChangeGC(dpy, gc, GCClipMask|GCClipXOrigin|GCClipYOrigin,&values);
    XCopyArea(dpy, logo, d, gc, 0, 0, LOGO_WIDTH, LOGO_HEIGHT, 0, 0);
    values.clip_mask = (int) NULL;
    XChangeGC(dpy, gc, GCClipMask|GCClipXOrigin|GCClipYOrigin,&values);
}

void DrawUI(Widget w, NPP instance, char *message)
{

    GC black_gc, white_gc;
    XGCValues values;
    PluginInstance *This;
    const char *id = "mplayerplug-in v" PACKAGE_VERSION;
    int id_width;
    int top, font_height, loops;
    int term, chop;

    This = (PluginInstance *) instance->pdata;
    if (0) {
	printf("Widget: %i\n", w);
	printf("Message: %s\n", message);
    }

    if (w == NULL)
	return;
    if (This->state < STATE_READY)
	return;
    if (message == NULL)
        return;
    if (DESTROYED == 1)
        return;
    if (This->display == NULL)
        return;

    sprintf(This->lastmessage,"%s",message);

    values.foreground =
	BlackPixel((Display *) This->display,
		   DefaultScreen((Display *) This->display));
    black_gc =
	XCreateGC((Display *) This->display, This->window, GCForeground,
		  &values);
    values.foreground =
	WhitePixel((Display *) This->display,
		   DefaultScreen((Display *) This->display));
    white_gc =
	XCreateGC((Display *) This->display, This->window, GCForeground,
		  &values);

    if (font == NULL) Initialize((Display *) This->display, This->window, black_gc);

    XFillRectangle((Display *) This->display, This->window, white_gc, 0, 0,
		   This->width, This->height);
    id_width = XTextWidth(font, id, strlen(id));
    if (This->height > 50 && This->width > id_width+BORDER*2) {
	top=BORDER*2+LOGO_HEIGHT;
	loops = 0;
        font_height = font->max_bounds.ascent + font->max_bounds.descent;
        DrawLogo((Display *) This->display, This->window, black_gc);
        XDrawString((Display *) This->display, This->window, black_gc,
		This->width-id_width-BORDER, This->height-font_height, id, strlen(id));
        // Loop a max of 10 times in case of problems (like too small of a width)
        while(strlen(message) != 0 && loops < 10) {
            chop = 0;
	    term = 0;
            while(XTextWidth(font, message, strlen(message) - chop) + BORDER*2 > This->width)
                chop++;
            if(strchr(message, '\n') != NULL &&
              strchr(message, '\n') - message < strlen(message) - chop) {
                chop = strlen(message) - (strchr(message, '\n') - message);
                term = 1;
            }
            XDrawString((Display *) This->display, This->window, black_gc, BORDER, top,
                message, strlen(message) - chop);
            message += strlen(message) - chop + term;
            top += font_height;
            loops++;
        }
    } else {
        XDrawString((Display *) This->display, This->window, black_gc, BORDER, 15,
		message, strlen(message));
    }

    XFreeGC((Display *) This->display, black_gc);
    XFreeGC((Display *) This->display, white_gc);

    XFlush((Display *) This->display);
    usleep(100);
}
